/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */





/************************************************************************
 * thrtab.c: implementation of thread table classes.
************************************************************************/





/************************************************************************
 * header files.
************************************************************************/
#include "threadP.h"





ThrTable::ThrTable( unsigned int item_count )
  : items( new Entry[item_count] ),
	nItems( item_count ),
	next_ndx( 0 )
{
	items = new Entry[nItems];
}


ThrTable::~ThrTable( void )
{
	delete[] items;
}



thread_t
ThrTable::Alloc( item_t type )
{
	unsigned start_ndx = next_ndx;
	unsigned n = start_ndx;


	if( items[n].type != item_t_unknown )
	{
		/* advance to next item and start linear search */
		n = (n+1) % nItems;
		while( n != start_ndx )
		{
			if( items[n].type == item_t_unknown )
			{
				/* we found an empty slot */
				break;
			}

			/* haven't found an empty slot yet - advance */
			n = (n+1) % nItems;
		}

		if( n == start_ndx )
		{
			/* we didn't find an empty slot */
			thr_report_error("*** thread table overflow [%d]***\n", nItems );
			abort();
		}
	}

	/* n now holds index of empty slot -
	 * claim the spot for an item of the given type */
	items[n].type = type;
	switch( type )
	{
	case item_t_thread:
		items[n].pval = &(items[n].tcb);
		break;

	case item_t_file:
		items[n].pval = &(items[n].file);
		break;

	case item_t_socket:
		items[n].pval = &(items[n].socket);
		break;

	case item_t_signal:
		items[n].pval = &(items[n].sig);
		break;

#if defined(i386_unknown_nt4_0)
	case item_t_wmsg:
		items[n].pval = &(items[n].wmsg);
		break;
#endif // defined(i386_unknown_nt4_0)

	default:
		thr_report_error("*** thread table allocation request for unknown type ***\n");
		abort();
	}

	/* update starting index for next time */
	next_ndx = ( n+1 ) % nItems;

	/* return tid associated with n */
	return ToTid( n );
}


void
ThrTable::Dealloc( thread_t tid )
{
	unsigned ndx = ToIndex( tid );

	// release the item
	items[ndx].pval->Destroy();
	items[ndx].type = item_t_unknown;
	items[ndx].pval = NULL;
}


thread_t
ThrTable::FindNextOwned( thread_t tid, item_t type ) const
{
	// try the first item of the given type
	thread_t ntid = FindNext( tid, type );
	thread_t ftid = ntid;

	// keep searching if we need to
	while( (ntid != THR_TID_UNSPEC) && 
			( ((tid == THR_TID_UNSPEC) && (ntid != ftid)) || 
			  ((tid != THR_TID_UNSPEC) && (ntid != tid)) ) &&
			!thr_curr->i_own( ntid ) )
	{
		// we did not find an item that current thread owns 
		// and we have not yet wrapped around - try again
		ntid = FindNext( ntid, type );
	}

	return ((ntid != THR_TID_UNSPEC) && thr_curr->i_own( ntid )) ? ntid : THR_TID_UNSPEC;
}


thread_t
ThrTable::FindNext( thread_t tid, item_t type ) const
{
	thread_t retTid = THR_TID_UNSPEC;


	if( tid == THR_TID_UNSPEC )
	{
		// the user is asking for the first item of the given type -
		// test item 1 for a match
		if( items[ToIndex(1)].type == type )
		{
			// tid 1 was the desired type
			retTid = 1;
		}
		else
		{
			// tid 1 was not the desired type, so continue searching from 1
			retTid = FindNext( 1, type );
		}
	}
	else
	{
		// find the next tid that represents the given type
		// after the given tid
		thread_t currtid = FindNext( tid );

		while( currtid != tid )
		{
			if( items[ToIndex(currtid)].type == type )
			{
				// we found an item of the correct type
				retTid = currtid;
				break;
			}
			currtid = FindNext( currtid );
		}

		// handle case were we wrapped entirely around,
		// and given tid *is* the next of the given type
		if( currtid == tid )
		{
			if( items[ToIndex(currtid)].type == type )
			{
				retTid = currtid;
			}
		}
	}

	return retTid;
}




/************************************************************************
 * void ThrTable::UnbindAll(void)
 *
 * Unbind all bindables from the current (dying) thread.
************************************************************************/

void
ThrTable::UnbindAll( void )
{
	unsigned i;


	assert( thr_curr != NULL );
	DO_TRACE_1(("ThrTable::UnbindAll(): unbinding all bindables for dying thread"));

	for( i = 1; i <= Size(); i++ )
	{
		if( thr_curr->i_own(i) )
		{
			thr_curr->Unbind(i);
		}
	}
}


#if defined(THREADLIB_DEBUG)
void
ThrTable::WriteTo( ostream& s ) const
{
	unsigned i;


    for (i = 0; i < nItems; i++)
	{
        if( items[i].type != item_t_unknown )
        {
            assert( items[i].pval != NULL );
		    items[i].pval->WriteTo( s );
        }
	}
}
#endif // defined(THREADLIB_DEBUG)


//------------------------------------------------------------------------------
// ThrTableIterator methods
//------------------------------------------------------------------------------

ThrTableIterator::ThrTableIterator( const ThrTable& thrtab )
  : tab( thrtab )
{
	Reset();
}


bool
ThrTableIterator::Next( thread_t* tidp )
{
	bool bOK = (curr_tid != THR_TID_UNSPEC);

	if( bOK )
	{
		if( tidp != NULL )
		{
			*tidp = curr_tid;
		}

		// advance to next item, watching for wraparound
		curr_tid = thr_tab->FindNext( curr_tid );
		if( curr_tid == 1 )
		{
			// we wrapped around - we're done
			curr_tid = THR_TID_UNSPEC;
		}
	}
	return bOK;
}



bool
ThrTableIterator::Next( item_t itype, thread_t* tidp )
{
	bool bFound = false;


	if( curr_tid != THR_TID_UNSPEC )
	{
		// if current item is of correct type, we found our item
		if( thr_tab->items[thr_tab->ToIndex(curr_tid)].type == itype )
		{
			bFound = true;
		}
		else
		{
			// current item isn't of the correct type - 
			// find the next one in the table of the indicated type
			thread_t old_tid = curr_tid;
			curr_tid = thr_tab->FindNext( old_tid, itype );

			// check whether we found an item, watching out for wraparound
			if( (curr_tid != THR_TID_UNSPEC) && (curr_tid > old_tid) )
			{
				bFound = true;
			}
		}

		if( bFound )
		{
			// return the tid of the item if desired
			if( tidp != NULL )
			{
				*tidp = curr_tid;
			}

			// advance to next item (not necessarily of the indicated type,
			// watching out for wraparound
			curr_tid = thr_tab->FindNext( curr_tid );
			if( curr_tid == 1 )
			{
				// we wrapped around, so we're done
				curr_tid = THR_TID_UNSPEC;
			}
		}
		else
		{
			// we didn't find anything interesting, so we're done
			curr_tid = THR_TID_UNSPEC;
		}
	}

	return bFound;
}






